/******************************************************************************
 * File: Node.cpp
 *
 * Note: See the Node.h file for comments on the class and its 
 * functions and variables.
 *
 * @author      Samuel Audet
 * @version     1.0, 2006-01-30
 *
 */
#include <stdlib.h>
#include <time.h>

#include <iostream>
#include <fstream>
using namespace std;

#include "Node.h"
#include "TransTable.h"

// if defined, will randomize a bit values from the heuristic evaluation
//#define RANDOMIZE

#define min(a,b)  ((a) > (b) ? (b) : (a))
#define max(a,b)  ((a) < (b) ? (b) : (a))

// *THE* transpositional table, think of it as a singleton...
TransTable transTable;


int Node::PROTECTED_WEIGHT[2]  = { 1, 1 };
int Node::DOUBLY_PROTECTED_WEIGHT[2] = { 1, 1 };
int Node::LINEUP_WEIGHT[2] = { 4, 4 };
int Node::ATTACK_WEIGHT[2] = { 3, 3 };

Node::Stack Node::theStack;
uint64_t Node::boardCodes[5][5][3] = {0};
uint64_t Node::playerCodes[2] = {0};
int Node::maximumDepth = 0;
time_t Node::startTime = 0;
bool Node::timesUp = 0;
bool Node::doEvaluation = true;

int Node::getSuccessors(Node *nodes[40]) {
    int n = 0;
    
    // if we have pieces on hand, put one
    if (piecesOnHand[player] > 0) {
        for (int i = 0; i < 5; i++) {
            for (int j = 0; j < 5; j++) {
                Node *node = placeNewPiece(i, j);
                if (node != NULL) 
                    nodes[n++] = node;
            }
        }
    } else {
        // otherwise, move pieces around. legal moves: N E S W
        const static int legalMoves[4][2] = {{-1, 0}, {0, 1}, {1, 0}, {0, -1}};
    
        for (int i = 0; i < 5; i++) {
            for (int j = 0; j < 5; j++) {
                if (board[i][j] == player) {
                    for (int k = 0; k < 4; k++) {
                        Node *node = movePiece(i, j, legalMoves[k][0], legalMoves[k][1]);
                        if (node != NULL) 
                            nodes[n++] = node;
                    }
                }
            }
        }
    }

    // make sure we never return more than 40 by mistake
    assert(n < 40);
    return n;
}

int Node::evaluateAs(Player p) {
    if (!doEvaluation) return 0;

    int value = 0;
    Player opponent = Opponent(p);    

    int rowCounts[5] = {0}, colCounts[5] = {0}, diagCounts[2] = {0};
    for (int i = 0; i < 5; i++) {
        for (int j = 0; j < 5; j++) {
            if (board[i][j] == p) {

                // check for north-south protection
                bool protectedNS = false;
                if (i == 0 || i == 4 || board[i-1][j] == p || board[i+1][j] == p) {
                    value += PROTECTED_WEIGHT[player];
                    protectedNS = true;
                }

                // check for east-west protection
                bool protectedEW = false;
                if (j == 0 || j == 4 || board[i][j-1] == p || board[i][j+1] == p) {
                    value += PROTECTED_WEIGHT[player];
                    protectedEW = true;
                }

                // check for ultimate double protection 
                if (protectedNS && protectedEW)
                    value += DOUBLY_PROTECTED_WEIGHT[player];

                // check if we are attacking anybody
                int attacks = 0;
                if (j < 3 && board[i][j+1] == opponent && board[i][j+2] == EMPTY)
                    attacks++;
                if (j > 1 && board[i][j-1] == opponent && board[i][j-2] == EMPTY)
                    attacks++;
                if (i < 3 && board[i+1][j] == opponent && board[i+2][j] == EMPTY)
                    attacks++;
                if (i > 1 && board[i-1][j] == opponent && board[i-2][j] == EMPTY)
                    attacks++;
                value += attacks*ATTACK_WEIGHT[player];

                // increment row and columns count
                rowCounts[i]++; colCounts[j]++;
            }
        }

        // increment diagonal counts
        if (board[i][i] == p) diagCounts[0]++;
        if (board[i][4-i] == p) diagCounts[1]++;
    }

    // add to value the counts squared (so that aligned pieces will get better values)
    for (int i = 0; i < 5; i++)
         value += LINEUP_WEIGHT[player] * (rowCounts[i]*rowCounts[i] + colCounts[i]*colCounts[i]);
    value += LINEUP_WEIGHT[player] * (diagCounts[0]*diagCounts[0] + diagCounts[1]*diagCounts[1]);

    value *= 10;
    assert(value > MINVALUE && value < MAXVALUE);
    return value;
}

int Node::minimax(int alpha, int beta, int depth, Node *maxNodeInOut, bool isGuessNode) {
    // check if we still have time after every 1024 node pops
    if (timesUp || (startTime != -1 && (theStack.popCount & 0x3FF) == 0 && 
            time(NULL) - startTime > TIME_PER_TURN-1)) {
        timesUp = true;
        return 0;
    }

    // if this board represents a win of the opponent -> leaf node and return -1
    if (hasLost()) {
        return MINVALUE;
    }

    TransTable::ValueType valueType;
    int value = transTable.get(this, maximumDepth - depth, valueType);
    if (value != 0) {
        // we have a value in transpositional table!
        switch (valueType) {
            case TransTable::EXACT_VALUE: return value;
            case TransTable::UPPER_BOUND: 
                if (value < alpha) 
                    return value;
                else if(value < beta)
                    beta = value;
                break;
            case TransTable::LOWER_BOUND: 
                if (value > beta) 
                    return value; 
                else if(value > alpha)
                    alpha = value;
                break;
            default: assert(false);
        }
    }

    if (depth >= maximumDepth) {
        // we reached search limit... call evaluation function
        int value = evaluateAs(player)-evaluateAs(Opponent(player));
#ifdef RANDOMIZE
        // randomize up to 10% of the value's size...  
        double random_factor = (0.9 + 0.2*random()/RAND_MAX);
        return lrint(value*random_factor);
#else
        return value;
#endif
    }

    int maxValue = MINVALUE-1;
    Node *maxNode = NULL;

    // if we already have a guess at the best node, try it first
    // to optimize alpha-beta pruning    
    if (depth == 0 && maxNodeInOut != NULL && isGuessNode) {
        alpha = maxValue = -maxNodeInOut->minimax(-beta, -alpha, depth+1);
        maxNode = maxNodeInOut;
    }

    Node *nodes[40];
    int n = getSuccessors(nodes);
    if (n > 0) {
        for (int i = 0; i < n; i++) {
            // the minus, and the inversion of beta and alpha are intentional:
            // the minimax() function is used as both min and max function, 
            // i.e.: negamax
            int value = -nodes[i]->minimax(-beta, -alpha, depth+1);
            if (value > maxValue) { 
                maxNode = nodes[i]; 
                maxValue = value;
            }

            if (maxValue >= beta) break;
            alpha = max(alpha, maxValue);
        }

        // If I am parent root node, return maxNode 
        if (depth == 0 && !timesUp) {
            *maxNodeInOut = *maxNode;
        }

        // push back the nodes on the stack
        for (int i = 0; i < n; i++)     
            theStack.push(nodes[i]);
    } else {
        // If we can't move at all, we lost
        maxValue = MINVALUE;
    }

    // put information in transpositional table
    if (maxValue < alpha) 
        valueType = TransTable::UPPER_BOUND;
    else if (maxValue >= beta) 
        valueType = TransTable::LOWER_BOUND;
    else 
        valueType = TransTable::EXACT_VALUE;
    transTable.put(this, maximumDepth - depth, valueType, maxValue);

    return maxValue;
}

int Node::play(int depth, bool doEvaluation, bool useTransTable) {
    // setup our stuff, and call minimax() with iterarive deepening
    this->doEvaluation = doEvaluation;
    transTable.clear();
    transTable.setEnabled(useTransTable);

    Node maxNode;
    int realvalue = 0;

    // if we have a fixed depth, use it, otherwise
    // set a startTime
    if (depth == -1)
        startTime = time(NULL);
    else
        startTime = -1;
    theStack.popCount = 0;
    timesUp = false;
    maximumDepth = 2;
    cout << "Ply: " << maximumDepth << " "; cout.flush();
    realvalue = minimax(MINVALUE-1, MAXVALUE+1, 0, &maxNode, false);
    // stop if we are out of time, but also if we already know a 
    // minimax path to the winning position
    while ((unsigned int) maximumDepth < (unsigned int) depth && !timesUp && 
           realvalue < MAXVALUE && realvalue > MINVALUE) {
        maximumDepth++;                         
        cout << maximumDepth << " "; cout.flush();
        int value = minimax(MINVALUE-1, MAXVALUE+1, 0, &maxNode, true);
        if (!timesUp) realvalue = value;

    }
    *this = maxNode;

    return realvalue;
}

void Node::print() {
    for (int i = 0; i < 5; i++) {
        for (int j = 0; j < 5; j++)
            cout << (board[i][j] == PLAYER0 ? '0' : board[i][j] == PLAYER1 ? '1' : ' ') << ',';
        cout << endl;
    }

    if (piecesOnHand[PLAYER0] > 0 || piecesOnHand[PLAYER1] > 0) {
        cout << "Pieces on hand: ";
        for (int i = 0; i < piecesOnHand[PLAYER0]; i++)
            cout << '0';
        cout << " ";
        for (int i = 0; i < piecesOnHand[PLAYER1]; i++)
            cout << '1';
        cout << endl;
    }
}

bool Node::parse() {
    // parse position where we want our piece
    cin.sync();
    if (piecesOnHand[player] > 0) {
        // if we have pieces on hand, do not parse N E S W, no directions
        char i, j;
        cin >> i >> j;
        i -= 49;
        j -= 49;

        Node *n = placeNewPiece(i, j);
        if (n == NULL) {
            cout << "Invalid move, (" << (int)i+1 << "," << (int)j+1 << ") is not empty, try again." << endl;
            return false;
        } else { 
            *this = *n;
            theStack.push(n);
            cout << endl;
            return true;
        }
    } else {
        char i, j, d;
        cin >> i >> j >> d;
        i -= 49;
        j -= 49;

        if (board[(int)i][(int)j] != player) {
            cout << "Invalid move, no opponent (you) piece at (" << (int)i+1 << "," << (int)j+1 << "), try again." << endl;
            return false;
        }

        int di = 0, dj = 0;
        switch (d) {
            case 'N': case 'n': di = -1; break;
            case 'S': case 's': di =  1; break;
            case 'E': case 'e': dj =  1; break;
            case 'W': case 'w': dj = -1; break;
            default: cout << d << ": Invalid direction, try again." << endl; return false;
        }

        Node *n = movePiece(i, j, di, dj);
        if (n == NULL) {
            cout << "Can't move in direction " << d << ", try again." << endl;      
            return false;
        } else { 
            *this = *n;
            theStack.push(n);           
            cout << endl;
            return true;
        }
    }
}


void Node::initCodes(bool readFile) {
    // initialize random stuff
    srandom(time(NULL));

    if (readFile) {
        ifstream file("random_numbers.txt");

        for (int i = 0; i < 5; i++)
            for (int j = 0; j < 5; j++)
                for (int k = 0; k < 3; k++) {
                    file >> boardCodes[i][j][k];
                }
        for (int i = 0; i < 2; i++) {
            file >> playerCodes[i];
        }
    } else {
        ofstream file("random_numbers.txt");

        for (int i = 0; i < 5; i++)
            for (int j = 0; j < 5; j++)
                for (int k = 0; k < 3; k++) {
                    boardCodes[i][j][k] = (uint64_t)random() | ((uint64_t)random()<<32);
                    file << boardCodes[i][j][k] << endl;
                }
        for (int i = 0; i < 2; i++) {
            playerCodes[i] = (uint64_t)random() | ((uint64_t)random()<<32);
            file << playerCodes[i] << endl;
        }
    }
}
