/******************************************************************************
 * File: Node.h
 */
#ifndef NODE_H
#define NODE_H

// number of seconds the computer is going to take to make a move
#define TIME_PER_TURN 7

// the Player enum and a function to find the Opponent of a player
enum Player { EMPTY = -1, PLAYER0 = 0, PLAYER1 = 1};
inline Player Opponent(Player player) {
    return (Player)(((int)player - 1) & 0x1);
}

/**
 * Node is the main class for the Black & White player. It incorporates
 * the minimax algorithm, alpha-beta pruning, the heuristic evaluation,
 * the Node stack, a "printer" and a "parser", Zobrist key computation, 
 * and all the logic to make it possible for the computer to make a move. 
 *
 * @author      Samuel Audet
 * @version     1.0, 2006-01-30
 */
class Node {
public:
    Node() {
        clear();
    }

    /** 
     * Clears the node to initial state
     *
     * @param player    the player to start the game, if not PLAYER1 = White
     */
    void clear(Player player = PLAYER1) {
        memset(board, EMPTY, sizeof(board));
        piecesOnHand[0] = piecesOnHand[1] = 5;
        this->player = player;
        computeKey();
        nextFree = NULL;
    }

    /** 
     * Returns a row of the board 
     *
     * @param i         row of the board
     */
    signed char *operator [](int i) { return board[i]; }
    /** 
     * Assignment operator, copies the Node 
     *
     * @param n         the node to copy from
     */
    Node &operator =(Node &n) { 
        memcpy(this->board, n.board, sizeof(board)); 
        memcpy(this->piecesOnHand, n.piecesOnHand, sizeof(piecesOnHand));
        player = n.player;
        key = n.key;
        return *this; 
    }
    /** 
     * Clones a node by creating a new one and copying its content
     *
     * @return          new node
     */
    Node *clone() {
        Node *myClone = theStack.pop();
        *myClone = *this;
        return myClone;
    }

    /** 
     * Creates a new node with piece at position (i,j)
     * moved by di and dj squares. Also updates Zobrist key.
     *
     * @return          new node, null if invalid move
     */
    Node *movePiece(int i, int j, int di, int dj) {
        int i2 = min(max(i+di,0),4);
        int j2 = min(max(j+dj,0),4);

        if (board[i2][j2] == EMPTY) {
	    // normal move
            Node *n = clone();
            (*n)[i ][j ] = EMPTY;
            (*n)[i2][j2] = player;
            n->key ^= boardCodes[i][j][player+1];
            n->key ^= boardCodes[i][j][EMPTY+1];            
            n->key ^= boardCodes[i2][j2][EMPTY+1];
            n->key ^= boardCodes[i2][j2][player+1];
            n->player = Opponent(player);
            n->key ^= playerCodes[player];
            n->key ^= playerCodes[Opponent(player)];
            return n;
        } else {
	    // capture move
            int i3 = min(max(i2+di,0),4);
            int j3 = min(max(j2+dj,0),4);
            if (board[i2][j2] == Opponent(player) && board[i3][j3] == EMPTY) {
                Node *n = clone();
                (*n)[i ][j ] = EMPTY;
                (*n)[i2][j2] = player;
                (*n)[i3][j3] = player;
                n->key ^= boardCodes[i][j][player+1];
                n->key ^= boardCodes[i][j][EMPTY+1];
                n->key ^= boardCodes[i2][j2][Opponent(player)+1];
                n->key ^= boardCodes[i2][j2][player+1];
                n->key ^= boardCodes[i3][j3][EMPTY+1];
                n->key ^= boardCodes[i3][j3][player+1];
                n->player = Opponent(player);
                n->key ^= playerCodes[player];
                n->key ^= playerCodes[Opponent(player)];
                return n;
            }
        }
        return NULL; // can't move there
    }

    /** 
     * Creates a new node with a new piece at position (i,j)
     * Also updates Zobrist key.
     *
     * @return          new node, null is invalid move
     */
    Node *placeNewPiece(int i, int j) {
        if (board[i][j] == EMPTY) {
            Node *n = clone();
            (*n)[i][j] = player;
            n->key ^= boardCodes[i][j][EMPTY+1];
            n->key ^= boardCodes[i][j][player+1];
            n->piecesOnHand[player]--;
            n->player = Opponent(player);
            n->key ^= playerCodes[player];
            n->key ^= playerCodes[Opponent(player)];
            return n;
        } else
            return NULL;
    }
    /** 
     * Computes if the current player has lost the game (terminal state)
     *
     * @return          true if current player has lost 
     */
    bool hasLost() {
        // The winner is the first player to line up five of her pieces, 
        // either horizontally, vertically, or diagonally, or else to 
        // make it impossible for the opponent to make a valid move.

        bool hasPieces = piecesOnHand[player] > 0;
        for (int i = 0; i < 5; i++) 
            for (int j = 0; j < 5; j++)
                if (board[i][j] == player) {
                    hasPieces = true;
                    break;
                }
        if (!hasPieces) return true;

        // what if we can't move pieces that we have ??

        Player opponent = Opponent(player);

        // check for complete rows
        for (int i = 0; i < 5; i++) {
            int j;
            for (j = 0; j < 5; j++) {
                if (board[i][j] != opponent) break;
            }
            if (j == 5) return true;
        }

        // check for complete columns
        for (int j = 0; j < 5; j++) {
            int i;      
            for (i = 0; i < 5; i++) {
                if (board[i][j] != opponent) break;
            }
            if (i == 5) return true;
        }

        // check for both diagonals
        int i;
        for (i = 0; i < 5; i++) 
            if (board[i][i] != opponent) break;
        if (i == 5) return true;

        for (i = 0; i < 5; i++) 
            if (board[i][4-i] != opponent) break;
        if (i == 5) return true;

        return false;
    }    

    /** 
     * Computes the legal successor board configurations to this node 
     *
     * @param   nodes   the return array of nodes (max 40)
     * @return          the number of nodes returned in "nodes"
     */
    int getSuccessors(Node *nodes[40]);
    /** 
     * Evaluation heuristic that returns a value of the current board configuration
     * value for the player passed as parameter.
     *
     * @param   player  the player to evaluation this configuration for
     * @return          the value of the board to this player 
     */
    int evaluateAs(Player player);
    /** 
     * The minimax algorithm, or actually negamax with alpha-beta pruning 
     *
     * @param   alpha   the alpha for this call (largest value from parents up till now)
     * @param   beta    the beta for this call (maximum allowable value from parent)
     * @param   depth   depth of recursion (depth of search tree)
     * @param   maxNodeInOut  at depth 0, takes a max node out, and in if isGuessNode == true
     * @param   isGuessNode   if true, takes maxNodeInOut as a guess at best node
     *
     * @return          the maximum value of the board to this player 
     */
    int minimax(int alpha, int beta, int depth, Node *maxNodeInOut = NULL, bool isGuessNode = false);
    /** 
     * Starts a turn for the computer by bootstrapping minimax with 
     * iterative deepening 
     *
     * @param   depth         maximum depth (fixed) or if -1, until TIME_PER_TURN is elapsed
     * @param   doEvaluation  if true, will use the heuristic evaluation (you want this) 
     * @param   useTransTable if true, will use the transpositional table
     *
     * @return          the maximum value of the board to this player 
     */
    int play(int depth = -1, bool doEvaluation = true, bool useTransTable = true);
    
    /**
     * Prints the board on the standard output.
     */
    void print();
    /**
     * Parses the board from the standard input.
     */
    bool parse();

    /**
     * Gets the Zobrist key of this Node.
     * @return          the Zobrit key
     */
    uint64_t getKey() { return key; }
    /**
     * Gets the current player of this Node 
     * @return          the current player 
     */
    Player getPlayer() { return player; }

    /**
     * The Stack class that keeps on a stack already created node for reuse. 
     */
    class Stack {
    public:
        Stack() : top(NULL), popCount(0), pushCount(0) { }
        
        /**
         * Pops a Node off the stack 
         * @return          a recycled or new Node
         */
        Node *pop() {
            popCount++;
            if (top == NULL)
                return new Node;
            else {
                Node *n = top;
                top = n->nextFree;
                n->nextFree = NULL;
                return n;
            }
        }

        /**
         * Pushes a Node on the stack for recycling
         * @param   n       a Node to recycle
         */
        void push(Node *n) {
            pushCount++;
            n->nextFree = top;
            top = n;
        }

    private:
        Node *top;     // the top of the stack
        int popCount;  // how many pops
        int pushCount; // how many pushes
        
        friend class Node;
    };

    static Stack theStack;

    /**
     * Initializes the codes needed for making Zobrist keys
     *
     * @param  readFile  read random_numbers.txt instead of writing it
     */
    static void initCodes(bool readFile = false);

    /**
     * The weights used in the heuristic evaluation function.
     * The sets of two are used during training.
     */
    static int PROTECTED_WEIGHT[2];
    static int DOUBLY_PROTECTED_WEIGHT[2];
    static int LINEUP_WEIGHT[2];
    static int ATTACK_WEIGHT[2];

private:
    signed char board[5][5]; // a digital represensation of the board
    int piecesOnHand[2];     // the number of pieces each player has on hand
    Player player;           // the next player to make a move

    Node *nextFree;          // the next free node (only when on theStack)
    uint64_t key;            // the Zobrist key of this Node

    static uint64_t boardCodes[5][5][3]; // the codes used for making Zobrist keys
    static uint64_t playerCodes[2];

    static int maximumDepth;  // the maximum depth of Node evaluation
    static time_t startTime;  // the time the computer started to make a move
    static bool timesUp;      // a flag that is set to true if time()-startTime > TIME_PER_TURN
    static bool doEvaluation; // indicates if we want to do heuristic evaluation or not

    /**
     * Computers the Zobrist key of this Node from scratch.
     */
    void computeKey() {
        key = 0;
        for (int i = 0; i < 5; i++) {
            for (int j = 0; j < 5; j++) {
                key ^= boardCodes[i][j][board[i][j]+1];
            }
        }
        key ^= playerCodes[player];
    }

};

#endif // NODE_H
