/******************************************************************************
 * File: TransTable.h
 */
#ifndef TRANSTABLE_H
#define TRANSTABLE_H

#include <math.h>

#define CAPACITY (1<<23)     // the capacity in # of entries (power of 2 needed)
#define MAXVALUE ((1<<15)-1) // the maximum value storable in hash table
#define MINVALUE (-MAXVALUE) // the minimum value storable in hash table

/**
 * Implementation of a transpositional table on a hash table with Zobrist  
 * keys. 
 *
 * @author      Samuel Audet
 * @version     1.0, 2006-01-30
 */
class TransTable {
public:
    enum ValueType { EXACT_VALUE = 1, LOWER_BOUND = 2, UPPER_BOUND = 3 };

    TransTable() {
        for (int i = 0; i < CAPACITY; i++)
            hashTable[i] = 0;
        enabled = true;
        collisions = 0;
        hits = 0;
    }

    /** 
     * Clears the hash table.
     */
    void clear() {
        int count = 0;
        for (int i = 0; i < CAPACITY; i++) {
            if (hashTable[i] != 0) {
                count++;
                hashTable[i] = 0;
            }
        }
        double usage = 100.0*count/CAPACITY;

        if (usage > 80) {
            cout << "WARNING!! Transpositional table was more than 80% full." << endl;
            exit(1);
        }
        collisions = 0;
        hits = 0;
    }

    /** 
     * Puts info about a Node in the hash table. 
     *
     * @param n         the Node whose (needed) information we want to store 
     * @param height    the "height" of the Node (how much depth this Node has)
     * @param type      the type of the value (EXACT_VALUE, UPPER_BOUND, LOWER_BOUND)
     * @param value     the actual value we want to keep in storage for this node
     */
    void put(Node *n, int height, ValueType type, int value) {
        assert(height <= 32);
        assert(value >= MINVALUE && value <= MAXVALUE);
        if (enabled) {
            unsigned int hash = n->getKey() & (CAPACITY-1);
            DATA data; data.raw = hashTable[hash];
            
            // if there is no collision or if the old data has a lower height,
            // put the (better) stuff in there
            if (data.raw == 0 || (unsigned int) height > data.height) {
                // use lower bits of key as hash, and store in data the upper 41 bits + height + value
                data.keypart = n->getKey()>>(64-41);
                data.height = height;
                data.valueType = type;
                data.value = value;
                hashTable[hash] = data.raw;
            }
        }
    }

    /** 
     * Retrieves info about a Node in the hash table. 
     *
     * @param n         the Node whose (needed) information we want to retrive 
     * @param height    the "height" of the Node (how much depth this Node has)
     * @param type      the type of the value (EXACT_VALUE, UPPER_BOUND, LOWER_BOUND)
     * @return          the actual value associated with this Node 
     */
    int get(Node *n, int height, ValueType &type) {
        assert(height <= 32);
        if (enabled) {
            unsigned int hash = n->getKey() & (CAPACITY-1);
            DATA data; data.raw = hashTable[hash];
            // return nothing if there is no data, or if the height
            // is greater than the height we stored..
            // we don't want crummy information, so return nothing
            // to force the callee to make a better evaluation
            if (data.raw == 0 || (unsigned int) height > data.height)
                return 0;
            if (data.keypart == (n->getKey()>>(64-41))) {
                hits++;
                type = (ValueType) data.valueType;
                return data.value;
            } else {
                collisions++;
            }
        }
        return 0;
    }

    /** 
     * Returns the number of Nodes in the hash table. 
     *
     * @return          number of Nodes in hash table 
     */
    int getSize() {
        int count = 0;
        for (int i = 0; i < CAPACITY; i++)
            if (hashTable[i] != 0) count++;
        return count;
    }

    /** 
     * Returns the usage in percent of this hash table. 
     * should never exceed ~80%
     *
     * @return          usage in percent of the hash table 
     */
    int getUsage() {
        return lrint(100.0*getSize()/CAPACITY);
    }

    /** 
     * Returns the hash code mean of stored Nodes. 
     * Should be ~CAPACITY/2 to be a uniform distribution
     * otherwise, need better keys.
     *
     * @return          the mean of hash codes 
     */
    int getMean() {
        double mean = 0;
        int count = 0;
        for (int i = 0; i < CAPACITY; i++)
            if (hashTable[i] != 0) { 
                mean += i;
                count++;
            }
        mean /= count;
        return lrint(mean);
    }

    /** 
     * Returns the standard deviation of the hash codes
     * of stored Nodes. Should be ~0.29*CAPACITY to be a 
     * uniform distribution otherwise, need better keys.
     *
     * @return          the mean of hash codes 
     */
    int getStdDev() {
        double mean = getMean();
        double variance = 0;
        int count = 0;
        for (int i = 0; i < CAPACITY; i++)
            if (hashTable[i] != 0) { 
                variance += (mean-i)*(mean-i);
                count++;
            }
        variance /= count;
        return lrint(sqrt(variance));
    }

    /** 
     * Enables or disables the transpositional table.
     *
     * @param b         true, enables it
     */
    void setEnabled(bool b) { enabled = b; }
    /** 
     * Returns enabled state of the transpositional table.
     *
     * @return          true if enabled
     */
    bool isEnabled() { return enabled; }
    
    /** 
     * Returns the number of hits (good) get() has received
     *
     * @return          number of hits 
     */
    int getHits() { return hits; }
    /** 
     * Returns the number of collisions (bad) get() has received
     *
     * @return          number of collisions 
     */
    int getCollisions() { return collisions; }

private:
    /** 
     * Union of a raw 64 bit number and a bit-field to easily access
     * parts of this number
     */
    union DATA {
        struct {
            uint64_t keypart : 41;        // higher part of the key
            unsigned int height : 5;      // height of the stored node info
            unsigned int valueType : 2;   // type of value stored (see ValueType enum)
            signed short value;           // the actual value of this node
        };
        uint64_t raw;               // raw data representing this info in the hash table
    };

    int64_t hashTable[CAPACITY];    // the hash table
    bool enabled;                   // enabled or not
    int hits;                       // collects number of hits (good) in get()
    int collisions;                 // collects number of collisions (bad) in get()
};

#endif // TRANSTABLE_H
